==> sv_v0.11.tgz: supervisor/compcmd.tcl <==
#!/bin/sh
# the next line restarts using tclsh\
exec tclsh "$0" "$@"

# Command service of the component (Tcl/Tk) program. Must be sourced in the
# component script. Starting of the component is done by calling
# "open_server port".
#
# Following procedures should implement the component functionalities:
# cmd_set, cmd_get, cmd_init, cmd_park, cmd_go_local, cmd_go_online,
# cmd_run, cmd_stop. They all may set the global array cmp_pvar and variable
# cmp_status upon execution and return quickly. If a long-term action is 
# started by one of this procedure, the cmp_status is set BUSY and cmp_pvar
# is set cmp_pvar(wait)=<execution maximal time in seconds>. Upon completion of
# this action, the variable final result is saved in the array cmp_pvar and then
# the cmp_status is modified (set READY, PARKED, ERFAT or whatever).
# Modification of cmp_status is sensed here and the final reply is sent.
#
# ATTENTION! the variables cmp_status and array cmp_pvars are created here
# below.

# Status: a link to respective (first) element of component parameters
#upvar 0 cmp_pvar(status) cmp_status
set cmp_status PARKED
# Only statuses READY BUSY LOCAL PARKED NOSTAR ERSYN ERANG ERFAT are allowed.
# Note that they all are in uppercase!

# The parameters made available by the component to the request of SV:
array set cmp_pvar {}

# Flag of client connected:
set connected 0

# Client channel identifier:
set connected_chanID 0

#
#...Organize new connection of a client and exit when disconnected not blocking
#... the event loop. Called automatically by socket -server command.
#
proc accept_connection { chanID addr port } {
	global connected connected_chanID
	puts "Client connected: chanID = $chanID, port = $port, addr = $addr"
	fconfigure $chanID -buffering line
	puts [fconfigure $chanID]
	set connected 1
	set connected_chanID $chanID
	fileevent $chanID readable get_command_and_reply
	vwait connected
	if {! $connected} {
		puts "Reset $chanID reading handler"
		fileevent $chanID readable {}
	} else {
		error "Unexpected error 1."
	}
}

source parsecmd.tcl
#
#... Read the arrived command, call its parser, call its executor and reply
#... the result based on global cmp_status variable.
#	
proc get_command_and_reply {} {
	global connected connected_chanID cmp_status
	if {! [eof $connected_chanID] && [gets $connected_chanID cmd]>0} {
		puts "->$cmd"
		# call a command parser which returns a list like:
		# { comid kwd param1 value1|"switch" param2 value2|"switch" ... }
		set parsed_command [parse_command $cmd]
		# call a command executor which accepts parsed command list (product of
		# parse_command proc.) and returns the reply list.
		set answer "[process_command $parsed_command]"
		puts $connected_chanID "$answer"
		# - flushes automatically after fconfigure ... -buffering line
		puts "Sent: <$answer>"
		# Determine if the replied answer contained the delayed execution
		# message "WAIT=...":
		if {[string match -nocase *wait=* $answer]} {
			set ci [lindex $answer 0]
	puts "Delayes execution..."
			# waiting is reported. Trace the cmp_status variable to change:
			while {[string match -nocase busy $cmp_status]} {vwait cmp_status}
			# cmp_status is no longer busy -> report completion
			set answer [compose_reply $ci]
			puts $connected_chanID "$answer"
	puts "Completed."
		}
	} else {
		puts "Client disconnected"
#		close $connected_chanID
		set connected 0
	}
}

#
#... Perform the requested actions based on the list as returned by
#... parse_command Immediately after starting action calls compose_reply
#... with some parameters and returns its resulting reply string. Before
#... calling the executors and after composing the immediate reply cmp_pvar
#... is cleared. The utility takes care about the current value of the
#... component status (cmp_status global) to decide whether to call the
#... command executor or not. The reply is composed based on the value of
#... cmp_status after the command execution call returns. The only
#... successful action the utility performs itself is reporting the status
#... on GET STATUS arrival.
# 
proc process_command lcmd {
	global cmp_pvar cmp_status

	set ci [ lindex $lcmd 0 ]
	set kwd [ lindex $lcmd 1]
	# Remove COMID and KWD from reply leaving the parameter=value's only:
	set lcmd [lreplace $lcmd 0 1]
		
	# Utilities called below set cmp_status glob.variable according to which we
	# acknowledge the command after switch command, and set their replied
	# parameters in global array cmd_pvar.
	# If in local state, then first go READY/BUSY:
	if {$cmp_status=="LOCAL"} cmd_go_online

	# flag of reporting status in reply: usually 1
	set needstatus 1
	# flag of rejected command execution:
	set rejected 0
	# reset the component parameters reported to SV:
	array unset cmp_pvar * 
	
	# save component status:
	set prev_cmp_status $cmp_status
	
	# force switch below to return ERSYN in cmp_status if COMID is invalid:
	if {[string is integer $ci]==0} {set kwd bad}
	switch $kwd {
		"GET"
			{
				# create the array parameters and their values=switch:
				array set in_pvar $lcmd
				# create the list of parameter names:
				set pl [array names in_pvar]
				# if "status" is among them - remove it and set needstatus
				if {[set i [lsearch $pl status]]<0} {
					# no status there:
					set needstatus 0
				} else {
					set needstatus 1
					set pl [lreplace $pl $i $i]
				}
				# check whether it contained status only:
				if {[llength $pl]>0} {
					if {$cmp_status=="BUSY"} {
						set rejected 1
					} else {
						if {[llength [info procs cmd_get]]} {
							# sets the global cmp_pvar array elements:
							cmd_get $pl
						} else { set cmp_status ERSYN}
					}
				}
				# else - status will be reported due to needstatus==1
			}
		"SET"
			-
		"INIT"
			{
				# sets the component param-s array {par1=val1, par2=val2...}
				# Before returning, sets cmp_status
				if {$cmp_status=="READY"||$cmp_status=="PARKED"} {
					if {$kwd=="SET"} {
						# set the component parameters from {p v p v} list:
						if {[llength [info procs cmd_set]]} {
							set needstatus 0
							cmd_set lcmd
						} else { set cmp_status ERSYN}
					} else {
						# initialize the component hardware and parameters:
						if {[llength [info procs cmd_init]]} {
							cmd_init
						} else { set cmp_status ERSYN}
					}
				} else {
					set rejected 1
				}
			}
		"FREE"
			{
				if {$cmp_status=="READY"} {
					# give the control to local console: 
					if {[llength [info procs cmd_go_local]]} {
						cmd_go_local
					} else { set cmp_status ERSYN }
				} else {
					set rejected 1
				}
			}
		"PARK"
			{
				switch $cmp_status {
					"READY"
						-
					"ERFAT"
						{
							# park the component hardware:
							if {[llength [info procs cmd_park]]} {
								cmd_park
							} else { set cmp_status ERSYN }
						}
					"PARKED"
						{
							# nothing to do: already parked
						}
					default
						{
							set rejected 1
						}
				}
			}
		"RUN"
			{
				if {$cmp_status=="READY"} {
					# run the action with list of {p v p v}
					cmd_run $lcmd
				} else {
					set rejected 1
				}
			}
		"STOP"
			{
				if {$cmp_status=="BUSY" && $lcmd=="now switch"} {
					# stop the action under way
					if {[llength [info procs cmd_stop]]} {
						cmd_stop
					} else { 
						set cmp_status ERSYN
					}
				}
				# else - do nothing
			}
		default
			{
				set cmp_status ERSYN
			}
	}
	set answer [compose_reply $ci $rejected $needstatus]
	# reset component status if was caused by bad command:
	if {$cmp_status=="ERSYN"||$cmp_status=="ERANG"} {
		set cmp_status $prev_cmp_status
	}
	return $answer
}

#
#... Create the report based on the results of command processing and
#... global component parameters for SV and status
#... Called from process_command and get_command_and_reply after completion of
#... a delayed command execution.
#
proc compose_reply {ci {rejected 0} {needstatus 1}} {
	global cmp_status cmp_pvar
	# determine the keyword of acknowledgment by cmp_status:
	if {$rejected || [string match -nocase ER* $cmp_status]} {
		set kwd ERROR
		set needstatus 1
	} else {
		set kwd OK
	}
	# compose the answer:
	if {$needstatus} {
		set cmp_pvar(status) $cmp_status
	}
	set answer "[compose_command "$ci $kwd [array get cmp_pvar]"]"
	# reset the component parameters reported to SV before possible repeated
	# call after delayed execution:
	array unset cmp_pvar * 
	return $answer
}

#  This procedure opens the server side connection and should be started last
# as entering in the events loop. Exits after setting the global variable
# stopvar with closing the server socket.
proc open_server {port} {
	global stopvar
	puts "Prepare to accept commands in socket port $port..."
	set mychan [ socket -server accept_connection $port ]
	puts "Mychan=$mychan"
	vwait stopvar
	puts "Close server channel $mychan."
	close $mychan
}

==> sv_v0.13.tgz: supervisor/compcmd.tcl <==
#!/bin/sh
# the next line restarts using tclsh\
exec tclsh "$0" "$@"

# Command service of the component (Tcl/Tk) program. Must be sourced in the
# component script. Starting of the component is done by calling
# "open_server port".
#
# Following procedures should implement the component functionalities:
# cmd_set, cmd_get, cmd_init, cmd_park, cmd_go_local, cmd_go_online,
# cmd_run, cmd_stop. They all may set the global array cmp_pvar and variable
# cmp_status upon execution and return quickly. If a long-term action is 
# started by one of this procedure, the cmp_status is set BUSY and cmp_pvar
# is set cmp_pvar(wait)=<execution maximal time in seconds>. Upon completion of
# this action, the variable final result is saved in the array cmp_pvar and then
# the cmp_status is modified (set READY, PARKED, ERFAT or whatever).
# Modification of cmp_status is sensed here and the final reply is sent.
#
# ATTENTION! the variables cmp_status and array cmp_pvars are created here
# below.

# Status: a link to respective (first) element of component parameters
#upvar 0 cmp_pvar(status) cmp_status
set cmp_status PARKED
# Only statuses READY BUSY LOCAL PARKED NOSTAR ERSYN ERANG ERFAT are allowed.
# Note that they all are in uppercase!

# The parameters made available by the component to the request of SV:
array set cmp_pvar {}

# Flag of client connected:
set connected 0

# Client channel identifier:
set connected_chanID 0

#
#...Organize new connection of a client and exit when disconnected not blocking
#... the event loop. Called automatically by socket -server command.
#
proc accept_connection { chanID addr port } {
	global connected connected_chanID
	puts "Client connected: chanID = $chanID, port = $port, addr = $addr"
	fconfigure $chanID -buffering line
	puts [fconfigure $chanID]
	set connected 1
	set connected_chanID $chanID
	fileevent $chanID readable get_command_and_reply
	vwait connected
	if {! $connected} {
#		puts "Reset $chanID reading handler"
	#	fileevent $chanID readable {}
	} else {
		error "Unexpected error 1."
	}
}

source parsecmd.tcl
#
#... Read the arrived command, call its parser, call its executor and reply
#... the result based on global cmp_status variable.
#	
proc get_command_and_reply {} {
	global connected connected_chanID cmp_status
	if {! [eof $connected_chanID] && [gets $connected_chanID cmd]>0} {
		puts "->$cmd"
		# call a command parser which returns a list like:
		# { comid kwd param1 value1|"switch" param2 value2|"switch" ... }
		set parsed_command [parse_command $cmd]
		# call a command executor which accepts parsed command list (product of
		# parse_command proc.) and returns the reply list.
		set answer "[process_command $parsed_command]"
		puts $connected_chanID "$answer"
		# - flushes automatically after fconfigure ... -buffering line
		puts "Sent: <$answer>"
		# Determine if the replied answer contained the delayed execution
		# message "WAIT=...":
		if {[string match -nocase *wait=* $answer]} {
			set ci [lindex $answer 0]
	puts "Delayes execution..."
			# waiting is reported. Trace the cmp_status variable to change:
			while {[string match -nocase busy $cmp_status]} {vwait cmp_status}
			# cmp_status is no longer busy -> report completion
			set answer [compose_reply $ci]
			puts $connected_chanID "$answer"
	puts "Completed."
		}
	} else {
		puts "Client disconnected"
		close $connected_chanID
		set connected 0
	}
}

#
#... Perform the requested actions based on the list as returned by
#... parse_command Immediately after starting action calls compose_reply
#... with some parameters and returns its resulting reply string. Before
#... calling the executors and after composing the immediate reply cmp_pvar
#... is cleared. The utility takes care about the current value of the
#... component status (cmp_status global) to decide whether to call the
#... command executor or not. The reply is composed based on the value of
#... cmp_status after the command execution call returns. The only
#... successful action the utility performs itself is reporting the status
#... on GET STATUS arrival.
# 
# here we organize the global par-val list for background accessing it from
# cmd_run or cmd_init:
#
set comp_cmd_lcmd [list]
proc process_command lcmd {
	global cmp_pvar cmp_status comp_cmd_lcmd

	set ci [ lindex $lcmd 0 ]
	set kwd [ lindex $lcmd 1]
	# Remove COMID and KWD from reply leaving the parameter=value's only:
	set comp_cmd_lcmd [lreplace $lcmd 0 1]
		
	# Utilities called below set cmp_status glob.variable according to which we
	# acknowledge the command after switch command, and set their replied
	# parameters in global array cmd_pvar.
	# If in local state, then first go READY/BUSY:
	if {$cmp_status=="LOCAL"} cmd_go_online

	# flag of reporting status in reply: usually 1
	set needstatus 1
	# flag of rejected command execution:
	set rejected 0
	# reset the component parameters reported to SV:
	array unset cmp_pvar * 
	
	# save component status:
	set prev_cmp_status $cmp_status
	
	# force switch below to return ERSYN in cmp_status if COMID is invalid:
	if {[string is integer $ci]==0} {set kwd bad}
	switch $kwd {
		"GET"
			{
				# create the array parameters and their values=switch:
				array set in_pvar $comp_cmd_lcmd
				# create the list of parameter names:
				set pl [array names in_pvar]
				# if "status" is among them - remove it and set needstatus
				if {[set i [lsearch $pl status]]<0} {
					# no status there:
					set needstatus 0
				} else {
					set needstatus 1
					set pl [lreplace $pl $i $i]
				}
				# check whether it contained status only:
				if {[llength [info procs cmd_get]]} {
					# sets the global cmp_pvar array elements:
					cmd_get $pl
				} else { set cmp_status ERSYN}
				# else - status will be reported due to needstatus==1
			}
		"SET"
			-
		"INIT"
			{
				# sets the component param-s array {par1=val1, par2=val2...}
				# Before returning, sets cmp_status
				if {$cmp_status!="BUSY"||$cmp_status!="ERFAT"} {
					if {$kwd=="SET"} {
						# set the component parameters from {p v p v} list:
						if {[llength [info procs cmd_set]]} {
							set needstatus 0
							cmd_set comp_cmd_lcmd
						} else { set cmp_status ERSYN}
					} else {
						# initialize the component hardware and parameters:
						if {[llength [info procs cmd_init]]} {
							cmd_init
						} else { set cmp_status ERSYN}
					}
				} else {
					set rejected 1
				}
			}
		"FREE"
			{
				if {$cmp_status=="READY"} {
					# give the control to local console: 
					if {[llength [info procs cmd_go_local]]} {
						cmd_go_local
					} else { set cmp_status ERSYN }
				} else {
					set rejected 1
				}
			}
		"PARK"
			{
				switch $cmp_status {
					"READY"
						-
					"ERFAT"
						{
							# park the component hardware:
							if {[llength [info procs cmd_park]]} {
								cmd_park
							} else { set cmp_status ERSYN }
						}
					"PARKED"
						{
							# nothing to do: already parked
						}
					default
						{
							set rejected 1
						}
				}
			}
		"RUN"
			{
				if {$cmp_status=="READY" || $cmp_status=="ERANG"} {
					# run the action with list of {p v p v}
					set cmp_status "BUSY"
					after 1 cmd_run comp_cmd_lcmd
					vwait cmp_pvar(wait)
					if {$cmp_pvar(wait)==0} {array unset cmp_pvar wait}
				} else {
					set rejected 1
				}
			}
		"STOP"
			{
				if {$cmp_status=="BUSY" && $comp_cmd_lcmd=="now switch"} {
					# stop the action under way
					if {[llength [info procs cmd_stop]]} {
						cmd_stop
					} else { 
						set cmp_status ERSYN
					}
				}
				# else - do nothing
			}
		default
			{
				set cmp_status ERSYN
			}
	}
	set answer [compose_reply $ci $rejected $needstatus]
	# reset component status if was caused by bad command:
	if {$cmp_status=="ERSYN"||$cmp_status=="ERANG"} {
		set cmp_status $prev_cmp_status
	}
	return $answer
}

#
#... Create the report based on the results of command processing and
#... global component parameters for SV and status
#... Called from process_command and get_command_and_reply after completion of
#... a delayed command execution.
#
proc compose_reply {ci {rejected 0} {needstatus 1}} {
	global cmp_status cmp_pvar
	# determine the keyword of acknowledgment by cmp_status:
	if {$rejected || [string match -nocase ER* $cmp_status] || [string \
		match -nocase NOSTAR $cmp_status]} {
		set kwd ERROR
		set needstatus 1
	} else {
		set kwd OK
	}
	# compose the answer:
	if {$needstatus} {
		set cmp_pvar(status) $cmp_status
	}
	set answer [compose_command "$ci $kwd [array get cmp_pvar]"]
	# reset the component parameters reported to SV before possible repeated
	# call after delayed execution:
	array unset cmp_pvar * 
	return $answer
}

#  This procedure opens the server side connection and should be started last
# as entering in the events loop. Exits after setting the global variable
# stopvar with closing the server socket.
proc open_server {port} {
	global stopvar
	puts "Prepare to accept commands in socket port $port..."
	set mychan [ socket -server accept_connection $port ]
	puts "Mychan=$mychan"
	vwait stopvar
	puts "Close server channel $mychan."
	close $mychan
}
